# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_reduce_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. [N, C, H, W] ---> [N, 1, H, W]
        2. [N, C, H, W] ---> [N, C, 1, W]
        3. [N, C, H, W] ---> [N, C, H, 1]
        4. [N, C, H, W] ---> [N, C, 1, 1]
        5. [N, C, H, W] ---> [N, 1, 1, 1]
    '''
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    output_shape = copy.deepcopy(operator.inputs[0].type.shape)
    params = operator.raw_operator.reduce

    from coremltools.proto.NeuralNetwork_pb2 import ReduceLayerParams as Params
    # Adjust C-axis
    if params.axis in [Params.CHW, Params.C]:
        output_shape[1] = 1
    # Adjust H-axis
    if params.axis in [Params.CHW, Params.HW, Params.H]:
        output_shape[2] = 1
    # Adjust W-axis
    if params.axis in [Params.CHW, Params.HW, Params.W]:
        output_shape[3] = 1

    operator.outputs[0].type.shape = output_shape


register_shape_calculator('reduce', calculate_reduce_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import unittest
import sys
import numpy
from pyspark.ml.feature import Bucketizer
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlBucketizer(SparkMlTestCase):

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_bucketizer(self):
        values = [(0.1,), (0.4,), (1.2,), (1.5,)]
        data = self.spark.createDataFrame(values, ["features"])
        model = Bucketizer(splits=[-float("inf"), 0.5, 1.4, float("inf")], inputCol="features", outputCol="buckets")

        feature_count = len(data.select('features').first())
        model_onnx = convert_sparkml(model, 'Sparkml Bucketizer', [
            ('features', FloatTensorType([None, feature_count]))
        ], target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.setHandleInvalid("error").transform(data)
        expected = predicted.select("buckets").toPandas().values.astype(numpy.float32)
        data_np = [data.toPandas().values.astype(numpy.float32)]
        paths = save_data_models(data_np, expected, model, model_onnx, basename="SparkmlBucketizer")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['buckets'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_concat
from ....common._registration import register_converter


def convert_concat(scope, operator, container):
    if operator.raw_operator.concat.sequenceConcat:
        axis = 0
    else:
        axis = 1

    apply_concat(scope, operator.input_full_names, operator.output_full_names, container,
                 operator_name=operator.full_name, axis=axis)


register_converter('concat', convert_concat)



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_split_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. [N, C] ---> [N', C]
        2. [N, C, H, W] ---> [N', C, H, W]
    '''
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=[1, None])
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    output_shape = copy.deepcopy(operator.inputs[0].type.shape)

    divided = output_shape[1] / operator.raw_operator.split.nOutputs
    if divided != int(divided):
        raise RuntimeError('Variable dimension along C-axis must be divisible by partition number')

    output_shape[1] = int(divided)

    for i in range(len(operator.outputs)):
        operator.outputs[i].type.shape = copy.deepcopy(output_shape)


register_shape_calculator('split', calculate_split_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._apply_operation import apply_sub
from ....common._registration import register_converter


def convert_preprocessing_mean_image(scope, operator, container):
    mean_tensor_name = scope.get_unique_variable_name(operator.full_name + '_mean')

    # We assume that the first input's shape is [N, C, H, W] so that the mean image's shape, [C, H, W], can
    # be inferred from the first input's shape.
    container.add_initializer(mean_tensor_name, onnx_proto.TensorProto.FLOAT,
                              operator.inputs[0].type.shape[1:], operator.raw_operator.meanImage)

    # We assume that the first input variable's shape is [N, C, H, W] while the mean image's shape is [C, H, W]. Thus,
    # broadcasting should be enabled starting with axis=1.
    apply_sub(scope, [operator.inputs[0].full_name, mean_tensor_name], operator.output_full_names, container,
              axis=1, broadcast=1)


register_converter('meanImagePreprocessor', convert_preprocessing_mean_image)



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._apply_operation import apply_matmul
from ...common._registration import register_converter, register_shape_calculator


def convert_sparkml_pca(scope, operator, container):
    op = operator.raw_operator
    pc_tensor = scope.get_unique_variable_name('pc_tensor')
    container.add_initializer(pc_tensor, onnx_proto.TensorProto.FLOAT,
                              [op.pc.numRows, op.pc.numCols], list(op.pc.toArray().flatten()))
    apply_matmul(scope, [operator.inputs[0].full_name, pc_tensor], operator.output_full_names, container)


register_converter('pyspark.ml.feature.PCAModel', convert_sparkml_pca)


def calculate_sparkml_pca_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])
    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType([N, operator.raw_operator.getOrDefault('k')])


register_shape_calculator('pyspark.ml.feature.PCAModel', calculate_sparkml_pca_output_shapes)



# SPDX-License-Identifier: Apache-2.0

# This file defines the interface of the converter internal object for callback,
# So the usage of the methods and properties list here will not be affected among the different versions.

from onnxconverter_common.interface import *  # noqa



# SPDX-License-Identifier: Apache-2.0

# Rather than using ONNX protobuf definition throughout our codebase, we import ONNX protobuf definition here so that
# we can conduct quick fixes by overwriting ONNX functions without changing any lines elsewhere.
from onnx import onnx_pb as onnx_proto  # noqa
from onnx import helper

# Overwrite the make_tensor defined in onnx.helper because of a bug (string tensor get assigned twice)
from onnx import mapping
from onnx.onnx_pb import TensorProto
from onnx.helper import split_complex_to_pairs


def _check_onnx_version():
    import pkg_resources
    min_required_version = pkg_resources.parse_version('1.0.1')
    current_version = pkg_resources.get_distribution('onnx').parsed_version
    assert current_version >= min_required_version, 'ONNXMLTools requires ONNX version 1.0.1 or a newer one'


_check_onnx_version()


def _make_tensor_fixed(name, data_type, dims, vals, raw=False):
    '''
    Make a TensorProto with specified arguments.  If raw is False, this
    function will choose the corresponding proto field to store the
    values based on data_type. If raw is True, use "raw_data" proto
    field to store the values, and values should be of type bytes in
    this case.
    '''
    tensor = TensorProto()
    tensor.data_type = data_type
    tensor.name = name

    if (data_type == TensorProto.COMPLEX64 or
            data_type == TensorProto.COMPLEX128):
        vals = split_complex_to_pairs(vals)
    if raw:
        tensor.raw_data = vals
    else:
        field = mapping.STORAGE_TENSOR_TYPE_TO_FIELD[
            mapping.TENSOR_TYPE_TO_STORAGE_TENSOR_TYPE[data_type]]
        getattr(tensor, field).extend(vals)

    tensor.dims.extend(dims)
    return tensor


helper.make_tensor = _make_tensor_fixed



# SPDX-License-Identifier: Apache-2.0

from uuid import uuid4
import onnx
from ..common.onnx_ex import get_maximum_opset_supported
from ..common._topology import convert_topology
from ._parse import parse_libsvm

# Invoke the registration of all our converters and shape calculators
from . import shape_calculators
from . import operator_converters


def convert(model, name=None, initial_types=None, doc_string='', target_opset=None,
            targeted_onnx=onnx.__version__, custom_conversion_functions=None, custom_shape_calculators=None):
    """
    :param model: a libsvm model
    :param initial_types: a python list. Each element is a tuple of a variable name and a type defined in data_types.py
    :param name: The name of the graph (type: GraphProto) in the produced ONNX model (type: ModelProto)
    :param doc_string: A string attached onto the produced ONNX model
    :param target_opset: number, for example, 7 for ONNX 1.2, and 8 for ONNX 1.3.
    :param targeted_onnx: A string (for example, '1.1.2' and '1.2') used to specify the targeted ONNX version of the
    produced model. If ONNXMLTools cannot find a compatible ONNX python package, an error may be thrown.
    :param custom_conversion_functions: a dictionary for specifying the user customized conversion function
    :param custom_shape_calculators: a dictionary for specifying the user customized shape calculator
    :return: An ONNX model (type: ModelProto) which is equivalent to the input scikit-learn model
    """
    if initial_types is None:
        raise ValueError('Initial types are required. See usage of convert(...) in \
                         onnxmltools.convert.libsvm.convert for details')

    if name is None:
        name = str(uuid4().hex)
    target_opset = target_opset if target_opset else get_maximum_opset_supported()

    # Parse scikit-learn model as our internal data structure (i.e., Topology)
    topology = parse_libsvm(model, initial_types, custom_conversion_functions,
                            custom_shape_calculators)

    # Infer variable shapes
    topology.compile()

    # Convert our Topology object into ONNX. The outcome is an ONNX model.
    onnx_model = convert_topology(topology, name, doc_string, target_opset, targeted_onnx)

    return onnx_model



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
from distutils.version import StrictVersion
import onnx
import pandas
import numpy
from pyspark.ml.classification import GBTClassifier
from pyspark.ml.linalg import Vectors
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase
from pyspark.ml.feature import StringIndexer


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmTreeEnsembleClassifier(SparkMlTestCase):

    @unittest.skipIf(sys.platform == 'win32',
                     reason="UnsatisfiedLinkError")
    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    @unittest.skipIf(StrictVersion(onnx.__version__) <= StrictVersion('1.3'), 'Need Greater Opset 9')
    def test_gbt_classifier(self):
        raw_data = self.spark.createDataFrame([
            (1.0, Vectors.dense(1.0)),
            (0.0, Vectors.sparse(1, [], []))
        ], ["label", "features"])
        string_indexer = StringIndexer(inputCol="label", outputCol="indexed")
        si_model = string_indexer.fit(raw_data)
        data = si_model.transform(raw_data)
        gbt = GBTClassifier(maxIter=5, maxDepth=2, labelCol="indexed", seed=42)
        model = gbt.fit(data)
        feature_count = data.first()[1].size
        model_onnx = convert_sparkml(model, 'Sparkml GBT Classifier', [
            ('features', FloatTensorType([None, feature_count]))
        ], spark_session=self.spark, target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(data)
        data_np = data.toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32),
            predicted.toPandas().probability.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        ]
        paths = save_data_models(data_np, expected, model, model_onnx,
                                    basename="SparkmlGBTClassifier")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['prediction', 'probability'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

'''
Utility functions for Spark ML to Onnx conversion intended for the end user mainly
'''
from ..common.data_types import StringTensorType, FloatTensorType


def buildInitialTypesSimple(dataframe):
    types = []
    for field in dataframe.schema.fields:
        types.append((field.name, getTensorTypeFromSpark(str(field.dataType))))
    return types


def getTensorTypeFromSpark(sparktype):
    if sparktype == 'StringType':
        return StringTensorType([1, 1])
    elif sparktype == 'DecimalType' \
            or sparktype == 'DoubleType' \
            or sparktype == 'FloatType' \
            or sparktype == 'LongType' \
            or sparktype == 'IntegerType' \
            or sparktype == 'ShortType' \
            or sparktype == 'ByteType' \
            or sparktype == 'BooleanType':
        return FloatTensorType([1, 1])
    else:
        raise TypeError("Cannot map this type to Onnx types: " + sparktype)


def buildInputDictSimple(dataframe):
    import numpy
    result = {}
    for field in dataframe.schema.fields:
        if str(field.dataType) == 'StringType':
            result[field.name] = dataframe.select(field.name).toPandas().values
        else:
            result[field.name] = dataframe.select(field.name).toPandas().values.astype(numpy.float32)
    return result


class SparkMlConversionError(Exception):
    pass



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_merge_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. [N, C_1, H_1, W_1], ..., [N, C_n, H_n, W_n] --->
            [N, max(C_1, ..., C_n), max(H_1, ..., H_n), max(W_1, ..., W_n)]

    If 'None' happens at any coordinate, that coordinate's final dimension would be 'None'.
    '''
    check_input_and_output_numbers(operator, input_count_range=[1, None], output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    # [TODO] Fix reduce-like shape inference. We now assume all inputs are 4-D.
    n_dims = max(len(variable.type.shape) for variable in operator.inputs)
    output_shape = [0] * n_dims
    for i in range(n_dims):
        input_dims = [variable.type.shape[i] for variable in operator.inputs if len(variable.type.shape) > i]
        if 'None' in input_dims:
            output_shape[i] = 'None'
        else:
            output_shape[i] = max(input_dims)

    operator.outputs[0].type.shape = output_shape


register_shape_calculator('add', calculate_merge_output_shapes)
register_shape_calculator('average', calculate_merge_output_shapes)
register_shape_calculator('max', calculate_merge_output_shapes)
register_shape_calculator('min', calculate_merge_output_shapes)
register_shape_calculator('multiply', calculate_merge_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import math
import numpy as np
from ....common._apply_operation import apply_affine, apply_mul, apply_div, apply_pad
from ....common._registration import register_converter


def calculate_legacy_pad_amount(H_in, pad_h, k_h, s_h):
    '''
    This function calculate padding amount along H-axis. It can be applied to other axes. It should be only used with
    pooling conversion.

    :param H_in: input dimension along H-axis
    :param pad_h: padding amount at H-axis
    :param k_h: kernel's H-axis dimension
    :param s_h: stride along H-axis
    :return: (top_padding_amount, bottom_padding_amount)
    '''
    # Calculate a common variable
    H_temp = H_in + 2 * pad_h - k_h
    # Pooling output shape under CoerML IncludeLastPixel padding mode
    H_include_last_pad_out = math.ceil(H_temp / s_h) + 1
    # Pooling output shape under valid padding mode
    H_valid_pad_out = math.floor(H_temp / s_h) + 1
    # Amount of values padded at top boundary. For max pooling, the padded value should be "-inf."
    # For average pooling, we should pad zeros.
    pad_t = pad_h
    # Amount of values padded at bottom boundary (add extra pixels so that H_include_last_pad_out = floor( (H_adjusted_out - k_h) / stride) + 1)
    if H_include_last_pad_out > H_valid_pad_out:
        pad_b = pad_h + (s_h - H_temp % s_h)
    else:
        pad_b = pad_h
    # Intermediate result with pad_t values at top and pad_b valules at bottom of the original input
    H_adjusted_out = H_in + pad_t + pad_b
    # Adjust padded result if the original pooling wants to cut off the last output pixel.
    if (H_include_last_pad_out - 1) * s_h >= H_in + pad_h:
        if H_adjusted_out % s_h == 0:
            H_adjusted_out -= s_h
        else:
            H_adjusted_out -= H_adjusted_out % s_h
    return (pad_t, H_adjusted_out - pad_t - H_in)


def create_legacy_pad(scope, input_name, output_name, H_in, W_in, k_h, k_w,
                      s_h, s_w, p_h, p_w, padded_value, container):
    '''
    This function adds one Pad operator into its last argument, which is a Container object. By feeding the output of
    the created Pad operator into Pool operator under valid padding mode, we can achieve the same functionality of
    CoreML' pooling under IncludeLastPixel padding mode.

    :param scope:
    :param input_name:
    :param output_name:
    :param H_in: input dimension along H-axis
    :param W_in: input dimension along W-axis
    :param k_h: kernel's H-axis dimension
    :param k_w: kernel's W-axis dimension
    :param s_h: stride along H-axis
    :param s_w: stride along W-axis
    :param p_h: padding amount at the beginning and the end of H-axis
    :param p_w: padding amount at the beginning and the end of W-axis
    :param padded_value: value used to fill padded area
    :param container: Container object
    '''
    # Add a Pad operator to pre-process 4-D tensor
    pad_t, pad_b = calculate_legacy_pad_amount(H_in, p_h, k_h, s_h)
    pad_l, pad_r = calculate_legacy_pad_amount(W_in, p_w, k_w, s_w)

    # CoreML pooling operator pads only their H- and W-axes. Here we assume the shape of the tensor to be padded
    # is [N, C, H, W], so we have 8 padding amounts
    #     pads = [N_begin_index, C_begin_index, H_begin_index, W_begin_index,
    #             N_end_index,   C_end_index,   H_end_index,   W_end_index]
    # Because only H- and W-axes are padded in CoreML, we leave padding amounts of N- and C-axes zeros.
    pads = [0, 0, pad_t, pad_l, 0, 0, pad_b, pad_r]
    apply_pad(scope, input_name, output_name, container, pads=pads, value=padded_value)


# The conversion of pooling has several possible outcomes. Let's first define some symbols and then discuss their
# ONNX computational graphs case-by-case.
#
# Symbols:
#  X: input 4-D tensor. It should have a shape [N, C, H, W] following CoreML's pooling definition.
#  Y: output tensor identical to CorML's pooling. Its shapes depends on the pooling type applied.
#
# Case 1: global pooling
#  X ---> ONNX Global Pooling ---> Y
# In this case, ONNX's pooling implementation should directly match CoreML's implementation, so it's just a naive
# translation.
#
# Case 2: local max/L2 pooling with same or valid padding
#
#  X ---> ONNX Local Max/L2 Pooling ---> Y
#
# In this case, ONNX's pooling implementation should directly match CoreML's implementation, so it's just a naive
# translation.
#
# Case 3: local max/L2 pooling under CoreML's IncludeLastPixel padding
#
#  X ---> Pad --> X' ---> ONNX Local Max/L2 Pooling ---> Y
#
# CoreML's IncludeLastPixel padding mode is not supported in ONNX's pooling. We combine a Pad
# operator and a pooling to simulate CoreML's behavior. In this case, the Pad takes all padding-related
# parameters from CoreML's pooling while ONNX's pooling is working under valid padding mode.
#
# Case 4: local average pooling with same or valid padding. exclude_pad_area is on.
#
#  X ---> ONNX Local Average Pooling ---> Y
#
# Current ONNX pooling operator just follows Caffe2, so the padded area is naturally excluded when calculating the
# numerator and denumerator of the pixel average covered by the kernel. That is, the translation from CoreML to ONNX
# is trivial.
#
# Case 5: local average pooling with same or valid padding. exclude_pad_area is off.
#
#  X ---> ONNX Local Average Pooling ---> Y' ------------> Mul ---> Y
#  |                                                        ^
#  |                                                        |
#  '---> Scaler ---> Z ---> ONNX L1-norm Pooling ---> Z' ---'
#
#  The Scaler has "alpha=0" and its "beta" is a constant. If "beta=1", the output of the L1-norm pooling, Z', is the
#  effective kernel size applied at each pixel when padded area is excluded. Here we use "beta=1/kerenel_size" so
#  that one value in Z' stands for
#         (the kernel size without padded area) / (the kernel size with padded area)
#  at a pixel. The output Y' is computed with exclude_pad_area=on, so the element-wise multiplication of Y' and Z'
#  is Y.
#
# Case 6: local average pooling with IncludeLastPixel padding. exclude_pad_area is on.
#
#  X ---> Pad ---> X' ---> ONNX Local Average Pooling ---> Y' ------> Div ---> Y
#  |                                                                                         ^
#  |                                                                                         |
#  '---> Scaler ---> Z ---> Pad ---> Z' ---> ONNX L1-norm Pooling ---> Z''
#
#  The Scaler has "alpha=0" and its "beta" is a constant. If "beta=1", the output of the L1-norm pooling, Z'', is
#  the effective kernel size applied at each pixel when padded area is excluded (since Pad fills the
#  padded area with zeros so those padded pixels are not counted by the L1-norm pooling). Here we use
#  "beta=1/kerenel_size" so that one value in Z' stands for
#         (the kernel size without padded area) / (the kernel size with padded area)
#  at a pixel. The output Y' is computed as if exclude_pad_area=on, so the element-wise division of Y' and Z'' is Y.
#
# Case 7: local average pooling with IncludeLastPixel padding. exclude_pad_area is off.
#
#  X ---> Pad --> X' ---> ONNX Local Average Pooling ---> Y
#
# Since Pad operators add zeros to X's margin and the local pooling here is working under valid padding, it's
# equivalent to the situation of exclude_pad_area=off.
def convert_pooling(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import PoolingLayerParams as Params
    from coremltools.proto.NeuralNetwork_pb2 import SamePadding

    params = operator.raw_operator.pooling
    # The input of Pool
    inputs = [variable.full_name for variable in operator.inputs]
    # The output of Pool
    outputs = [variable.full_name for variable in operator.outputs]

    # Handle global pooling mode. This case if much simpler than the conversion of local pooling.
    attrs = {'name': operator.full_name}
    if params.globalPooling:
        pooling_table = {params.MAX: 'GlobalMaxPool',
                         Params.AVERAGE: 'GlobalAveragePool',
                         Params.L2: 'GlobalLpPool'}

        if params.type not in pooling_table:
            raise ValueError('Unsupported pooling type: {}'.format(params.type))

        op_type = pooling_table[params.type]
        if params.type == Params.L2:
            attrs['p'] = 2
            container.add_node(op_type, inputs, outputs, op_version=2, **attrs)
        else:
            container.add_node(op_type, inputs, outputs, **attrs)
        return

    # From here to the end of this function, we will handle local pooling mode
    if params.type == Params.MAX:
        op_type = 'MaxPool'
        if container.target_opset < 8:
            op_version = 1
        elif container.target_opset < 10:
            op_version = 8
        else:
            op_version = 10
            attrs['ceil_mode'] = 0
    elif params.type == Params.AVERAGE:
        op_type = 'AveragePool'
        if container.target_opset < 7:
            op_version = 1
        elif container.target_opset < 10:
            op_version = 7
        else:
            op_version = 10
            attrs['ceil_mode'] = 0
    elif params.type == Params.L2:
        op_type = 'LpPool'
        attrs['p'] = 2
        op_version = 2
    else:
        raise ValueError('Unsupported pooling type: {}'.format(params.type))

    # CoreML default v.s. non-default parameters
    kernel_shape = [3, 3] if len(params.kernelSize) <= 0 else params.kernelSize
    strides = [1, 1] if len(params.stride) <= 0 else params.stride
    attrs['kernel_shape'] = kernel_shape
    attrs['strides'] = strides

    # Set up padding attributes
    pads = None
    auto_pad = None
    pad_type = params.WhichOneof('PoolingPaddingType')
    if pad_type == 'valid':
        if len(params.valid.paddingAmounts.borderAmounts) > 0:
            pads = [0, 0, 0, 0]
            pads[0] = params.valid.paddingAmounts.borderAmounts[0].startEdgeSize
            pads[1] = params.valid.paddingAmounts.borderAmounts[1].startEdgeSize
            pads[2] = params.valid.paddingAmounts.borderAmounts[0].endEdgeSize
            pads[3] = params.valid.paddingAmounts.borderAmounts[1].endEdgeSize
            # If padding amounts are all zero, there should be no padding list.
            if all(pad == 0 for pad in pads):
                pads = None
                auto_pad = 'VALID'
        else:
            auto_pad = 'VALID'
    elif pad_type == 'same':
        if params.same.asymmetryMode == SamePadding.BOTTOM_RIGHT_HEAVY:
            # In CoreML, BOTTOM_RIGHT_HEAVY means that the extra pixel (when not dividable by 2) will be added to the
            # end of an axis. This behavior matches ONNX's SAME_UPPER mode.
            # Reference: https://apple.github.io/coremltools/coremlspecification/sections/NeuralNetwork.html#samepadding
            #            https://github.com/onnx/onnx/blob/rel-1.2.1/docs/Operators.md#AveragePool
            auto_pad = 'SAME_UPPER'
        elif params.same.asymmetryMode == SamePadding.TOP_LEFT_HEAVY:
            # In CoreML, TOP_LEFT_HEAVY means that the extra pixel (when not dividable by 2) will be added to the
            # beginning of an axis. This behavior matches ONNX's SAME_LOWER mode.
            # Reference: https://apple.github.io/coremltools/coremlspecification/sections/NeuralNetwork.html#samepadding
            #            https://github.com/onnx/onnx/blob/rel-1.2.1/docs/Operators.md#AveragePool
            auto_pad = 'SAME_LOWER'
        else:
            raise ValueError('Unknown asymmetric mode: {}'.format(params.same.asymmetryMode))
    elif pad_type == 'includeLastPixel':
        # Here we use a Pad operator to mimic the behavior of this CoreML padding.
        auto_pad = 'VALID'
        H = operator.inputs[0].type.shape[2]
        W = operator.inputs[0].type.shape[3]
        pad_h = params.includeLastPixel.paddingAmounts[0]
        pad_w = params.includeLastPixel.paddingAmounts[1]
        legacy_padded_tensor_name = scope.get_unique_variable_name('legacy_padded_tensor')
        padded_value = 0. if params.type != Params.MAX else 1 + np.finfo(np.float32).min
        # Create a sub-graph of cases 3, 6, 7: X ---> Pad ---> X'
        create_legacy_pad(scope, inputs[0], legacy_padded_tensor_name, H, W, kernel_shape[0], kernel_shape[1],
                          strides[0], strides[1], pad_h, pad_w, padded_value, container)
        # Set the first input name to the output of Pad so that the following Pool operator won't access the
        # original input.
        inputs[0] = legacy_padded_tensor_name
    else:
        raise ValueError('Unsupported padding mode: {}'.format(pad_type))

    if pads is not None:
        attrs['pads'] = pads
    if auto_pad is not None:
        attrs['auto_pad'] = auto_pad

    if (params.type == Params.AVERAGE and params.avgPoolExcludePadding and pad_type == 'includeLastPixel') or \
            (params.type == Params.AVERAGE and not params.avgPoolExcludePadding and pad_type != 'includeLastPixel'):
        # Case 5 & 6. See comment above.

        # X --> Affine --> Z
        X_name = operator.inputs[0].full_name
        Y_name = operator.outputs[0].full_name
        Z_name = scope.get_unique_variable_name('Z')
        apply_affine(scope, X_name, Z_name, container, alpha=0., beta=1. / (kernel_shape[0] * kernel_shape[1]))

        Z_prime_name = scope.get_unique_variable_name('Z_prime')
        Y_prime_name = scope.get_unique_variable_name('Y_prime')

        if pad_type != 'includeLastPixel':
            # Create the major Pool operator.
            # Associated sub-graph of case 5: X ---> Pool ---> Y'
            container.add_node(op_type, inputs, Y_prime_name, **attrs)

            # Create operators to calculate correction coefficients
            # Associated sub-graph of case 5: Z ---> L1Pool ---> Z'
            lp_pool_attrs = {'name': scope.get_unique_operator_name('LpPool'), 'kernel_shape': kernel_shape,
                             'strides': strides, 'p': 1}
            if pads is not None:
                lp_pool_attrs['pads'] = pads
            if auto_pad is not None:
                lp_pool_attrs['auto_pad'] = auto_pad
            container.add_node('LpPool', Z_name, Z_prime_name, op_version=2, **lp_pool_attrs)

            # Element-wisely apply adjustment coefficients and create the expected CoreML output
            # Associated sub-graph of case 5: Y', Z' ---> Mul ---> Y
            apply_mul(scope, [Y_prime_name, Z_prime_name], Y_name, container, broadcast=0)
        else:
            # Create the major Pool operator
            # Associated sub-graph of case 6: X' ---> Pool ---> Y'
            container.add_node(op_type, inputs, Y_prime_name, **attrs)

            # Create operators to correct Pool's output
            Y_name = operator.outputs[0].full_name
            Z_prime_prime_name = scope.get_unique_variable_name('Z_prime_prime')

            # Pad the constant tensor.
            # Associated sub-graph of case 6: Z ---> Pad ---> Z'
            create_legacy_pad(scope, Z_name, Z_prime_name, operator.inputs[0].type.shape[2],
                              operator.inputs[0].type.shape[3], kernel_shape[0], kernel_shape[1],
                              strides[0], strides[1], params.includeLastPixel.paddingAmounts[0],
                              params.includeLastPixel.paddingAmounts[1], 0., container)

            # Associated sub-graph of case 6: Z' ---> L1Pool ---> Z''
            lp_pool_attrs = {'name': scope.get_unique_operator_name('LpPool'), 'kernel_shape': kernel_shape,
                             'strides': strides, 'p': 1, 'auto_pad': 'VALID'}
            container.add_node('LpPool', Z_prime_name, Z_prime_prime_name, op_version=2, **lp_pool_attrs)

            # Element-wisely apply adjustment coefficients and create the expected CoreML output
            # Associated sub-graph of case 6: Y', Z''  ---> Div ---> Y
            apply_div(scope, [Y_prime_name, Z_prime_prime_name], Y_name, container, broadcast=0)
    else:
        # Create the major Pool operator
        container.add_node(op_type, inputs, outputs, op_version=op_version, **attrs)


register_converter('pooling', convert_pooling)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter, register_shape_calculator
from ....proto import onnx_proto
from ...common.data_types import FloatTensorType


def convert_sparkml_one_hot_encoder(scope, operator, container):
    op = operator.raw_operator
    C = operator.inputs[0].type.shape[1]

    # encoded_slot_sizes[i] is the number of output coordinates associated with the ith categorical feature

    # Variable names produced by one-hot encoders. Each of them is the encoding result of a categorical feature.
    final_variable_names = []
    final_variable_lengths = []
    for i in range(0, len(op.categorySizes)):
        catSize = op.categorySizes[i]
        cats = range(0, catSize)
        # Put a feature index we want to encode to a tensor
        index_variable_name = scope.get_unique_variable_name('target_index')
        container.add_initializer(index_variable_name, onnx_proto.TensorProto.INT64, [1], [i])

        # Extract the categorical feature from the original input tensor
        extracted_feature_name = scope.get_unique_variable_name('extracted_feature_at_' + str(i))
        extractor_type = 'ArrayFeatureExtractor'
        extractor_attrs = {'name': scope.get_unique_operator_name(extractor_type)}
        container.add_node(extractor_type, [operator.inputs[0].full_name, index_variable_name],
                           extracted_feature_name, op_domain='ai.onnx.ml', **extractor_attrs)

        # Encode the extracted categorical feature as a one-hot vector
        encoder_type = 'OneHotEncoder'
        encoder_attrs = {'name': scope.get_unique_operator_name(encoder_type), 'cats_int64s': cats}
        encoded_feature_name = scope.get_unique_variable_name('encoded_feature_at_' + str(i))
        container.add_node(encoder_type, extracted_feature_name, encoded_feature_name, op_domain='ai.onnx.ml',
                           **encoder_attrs)

        # Collect features produce by one-hot encoders
        final_variable_names.append(encoded_feature_name)
        # For each categorical value, the length of its encoded result is the number of all possible categorical values
        final_variable_lengths.append(catSize)

    # Combine encoded features and passed features
    collector_type = 'FeatureVectorizer'
    collector_attrs = {'name': scope.get_unique_operator_name(collector_type)}
    collector_attrs['inputdimensions'] = final_variable_lengths
    container.add_node(collector_type, final_variable_names, operator.outputs[0].full_name,
                       op_domain='ai.onnx.ml', **collector_attrs)


register_converter('pyspark.ml.feature.OneHotEncoderModel', convert_sparkml_one_hot_encoder)


def calculate_sparkml_one_hot_encoder_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. [N, C] ---> [N, C']
        2. [N, 'None'] ---> [N, 'None']
    '''
    op = operator.raw_operator

    # encoded_slot_sizes[i] is the number of output coordinates associated with the ith categorical feature.
    encoded_slot_sizes = op.categorySizes

    N = operator.inputs[0].type.shape[0]
    # Calculate the output feature length by replacing the count of categorical
    # features with their encoded widths
    if operator.inputs[0].type.shape[1] != 'None':
        C = operator.inputs[0].type.shape[1] - 1 + sum(encoded_slot_sizes)
    else:
        C = 'None'

    operator.outputs[0].type = FloatTensorType([N, C])


register_shape_calculator('pyspark.ml.feature.OneHotEncoderModel', calculate_sparkml_one_hot_encoder_output_shapes)




# SPDX-License-Identifier: Apache-2.0

import sys
import inspect
import unittest
from distutils.version import StrictVersion
import onnx
import pandas
import numpy
from pyspark.ml import Pipeline
from pyspark.ml.classification import DecisionTreeClassifier
from pyspark.ml.linalg import VectorUDT, SparseVector, Vectors
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import StringTensorType, FloatTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, compare_results, run_onnx_model
from tests.sparkml import SparkMlTestCase
from pyspark.ml.feature import StringIndexer, VectorIndexer


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmDecisionTreeClassifier(SparkMlTestCase):

    @unittest.skipIf(True, reason="Mismatched input dimensions.")
    @unittest.skipIf(sys.platform == 'win32',
                     reason="UnsatisfiedLinkError")
    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    @unittest.skipIf(StrictVersion(onnx.__version__) <= StrictVersion('1.3'), 'Need Greater Opset 9')
    def test_tree_pipeline(self):
        import os
        this_script_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        input_path = os.path.join(this_script_dir, "data", "sample_libsvm_data.txt")
        original_data = self.spark.read.format("libsvm").load(input_path)
        #
        # truncate the features
        #
        feature_count = 5
        self.spark.udf.register("truncateFeatures",
                                lambda x: SparseVector(feature_count, range(0, feature_count), x.toArray()[125:130]),
                                VectorUDT())
        data = original_data.selectExpr("cast(label as string) as label", "truncateFeatures(features) as features")
        label_indexer = StringIndexer(inputCol="label", outputCol="indexedLabel", handleInvalid='error')
        feature_indexer = VectorIndexer(inputCol="features", outputCol="indexedFeatures",
                                        maxCategories=10, handleInvalid='error')

        dt = DecisionTreeClassifier(labelCol="indexedLabel", featuresCol="indexedFeatures")
        pipeline = Pipeline(stages=[label_indexer, feature_indexer, dt])
        model = pipeline.fit(data)
        model_onnx = convert_sparkml(model, 'Sparkml Decision Tree Pipeline', [
            ('label', StringTensorType([None, 1])),
            ('features', FloatTensorType([None, feature_count]))
        ], spark_session=self.spark, target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(data.limit(1))
        data_np = {
            'label': data.limit(1).toPandas().label.values.reshape((-1, 1)),
            'features': data.limit(1).toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        }
        expected = [
            predicted.toPandas().indexedLabel.values.astype(numpy.int64),
            predicted.toPandas().prediction.values.astype(numpy.int64),
            predicted.toPandas().probability.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        ]
        paths = save_data_models(data_np, expected, model, model_onnx,
                                basename="SparkmlDecisionTreePipeline")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['indexedLabel', 'prediction', 'probability'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(True, reason="Mismatched input dimensions.")
    @unittest.skipIf(sys.platform == 'win32',
                     reason="UnsatisfiedLinkError")
    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_tree_one_class_classification(self):
        features = [[0., 1.], [1., 1.], [2., 0.]]
        features = numpy.array(features, dtype=numpy.float32)
        labels = [1, 1, 1]
        dd = [(labels[i], Vectors.dense(features[i])) for i in range(len(labels))]
        data = self.spark.createDataFrame(self.spark.sparkContext.parallelize(dd), schema=["label", "features"])
        dt = DecisionTreeClassifier(labelCol="label", featuresCol="features")
        model = dt.fit(data)
        model_onnx = convert_sparkml(model, 'Sparkml Decision Tree One Class', [
            ('features', FloatTensorType([None, 2]))
        ], spark_session=self.spark, target_opset=TARGET_OPSET)
        data_np = data.toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        predicted = model.transform(data)
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32),
            predicted.toPandas().probability.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        ]
        paths = save_data_models(data_np, expected, model, model_onnx,
                                basename="SparkmlDecisionTreeBinaryClass")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['prediction', 'probability'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(sys.platform == 'win32',
                     reason="UnsatisfiedLinkError")
    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_tree_binary_classification(self):
        features = [[0, 1], [1, 1], [2, 0]]
        features = numpy.array(features, dtype=numpy.float32)
        labels = [0, 1, 0]
        dd = [(labels[i], Vectors.dense(features[i])) for i in range(len(labels))]
        data = self.spark.createDataFrame(self.spark.sparkContext.parallelize(dd), schema=["label", "features"])
        dt = DecisionTreeClassifier(labelCol="label", featuresCol="features")
        model = dt.fit(data)
        model_onnx = convert_sparkml(model, 'Sparkml Decision Tree Binary Class', [
            ('features', FloatTensorType([None, 2]))
        ], spark_session=self.spark, target_opset=TARGET_OPSET)
        data_np = data.toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        predicted = model.transform(data)
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32),
            predicted.toPandas().probability.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        ]
        paths = save_data_models(data_np, expected, model, model_onnx,
                                basename="SparkmlDecisionTreeBinaryClass")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['prediction', 'probability'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(sys.platform == 'win32',
                     reason="UnsatisfiedLinkError")
    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_tree_multiple_classification(self):
        features = [[0, 1], [1, 1], [2, 0], [0.5, 0.5], [1.1, 1.1], [2.1, 0.1]]
        features = numpy.array(features, dtype=numpy.float32)
        labels = [0, 1, 2, 1, 1, 2]
        dd = [(labels[i], Vectors.dense(features[i])) for i in range(len(labels))]
        data = self.spark.createDataFrame(self.spark.sparkContext.parallelize(dd), schema=["label", "features"])
        dt = DecisionTreeClassifier(labelCol="label", featuresCol="features")
        model = dt.fit(data)
        model_onnx = convert_sparkml(model, 'Sparkml Decision Tree Multi Class', [
            ('features', FloatTensorType([None, 2]))
        ], spark_session=self.spark, target_opset=TARGET_OPSET)
        data_np = data.toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        predicted = model.transform(data)
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32),
            predicted.toPandas().probability.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        ]
        paths = save_data_models(data_np, expected, model, model_onnx,
                                basename="SparkmlDecisionTreeMultiClass")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['prediction', 'probability'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
from distutils.version import StrictVersion
import numpy
import pandas
import onnx
from onnx.defs import onnx_opset_version
from pyspark.ml.feature import VectorIndexer
from pyspark.ml.linalg import Vectors
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlVectorIndexer(SparkMlTestCase):

    @unittest.skipIf(
        True, reason=(
            "discrepency, unfound values are replaced by -1 by ONNX and 0 "
            "by spark."))
    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    @unittest.skipIf(StrictVersion(onnx.__version__) <= StrictVersion('1.3'),
                     'Need Greater Opset 9')
    def test_model_vector_indexer_multi(self):
        vi = VectorIndexer(maxCategories=2, inputCol="a", outputCol="indexed")
        data = self.spark.createDataFrame([
            (Vectors.dense([-1.0, 1.0, 3.1]),),
            (Vectors.dense([0.0, 5.0, 6.2]),),
            (Vectors.dense([0.0, 9.0, 3.1]),)],
            ["a"]
        )
        model = vi.fit(data)
        model_onnx = convert_sparkml(model, 'Sparkml VectorIndexer Multi',  [
            ('a', FloatTensorType([None, model.numFeatures]))
        ], target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(data)
        expected = predicted.toPandas().indexed.apply(lambda x: pandas.Series(x.toArray())).values
        data_np = data.toPandas().a.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        paths = save_data_models(data_np, expected, model, model_onnx,
                                    basename="SparkmlVectorIndexerMulti")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['indexed'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    @unittest.skipIf(StrictVersion(onnx.__version__) <= StrictVersion('1.3'),
                     'Need Greater Opset 9')
    def test_model_vector_indexer_single(self):
        vi = VectorIndexer(maxCategories=3, inputCol="a", outputCol="indexed")
        data = self.spark.createDataFrame([
            (Vectors.dense([-1.0]),),
            (Vectors.dense([0.0]),),
            (Vectors.dense([0.0]),)],
            ["a"]
        )
        model = vi.fit(data)
        model_onnx = convert_sparkml(model, 'Sparkml VectorIndexer Single',  [
            ('a', FloatTensorType([None, model.numFeatures]))
        ], target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(data)
        expected = predicted.toPandas().indexed.apply(lambda x: pandas.Series(x.toArray())).values
        data_np = data.toPandas().a.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        paths = save_data_models(data_np, expected, model, model_onnx,
                                    basename="SparkmlVectorIndexerSingle")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['indexed'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import pandas
import numpy
from pyspark.ml.feature import MinHashLSH
from pyspark.ml.linalg import Vectors
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmMinHashLSH(SparkMlTestCase):

    @unittest.skipIf(True, reason="Discrepencies (Float -> Double?).")
    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_min_hash_lsh(self):
        data = self.spark.createDataFrame([
            (0, Vectors.sparse(6, [0, 1, 2], [1.0, 1.0, 1.0]),),
            (1, Vectors.sparse(6, [2, 3, 4], [1.0, 1.0, 1.0]),),
            (2, Vectors.sparse(6, [0, 2, 4], [1.0, 1.0, 1.0]),)
        ], ["id", "features"])
        mh = MinHashLSH(inputCol="features", outputCol="hashes", numHashTables=5)
        model = mh.fit(data)

        feature_count = data.first()[1].size
        model_onnx = convert_sparkml(model, 'Sparkml MinHashLSH', [
            ('features', FloatTensorType([None, feature_count]))
        ], spark_session=self.spark, target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data.limit(2))
        data_np = data.limit(2).toPandas().features.apply(
            lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        expected = [
            predicted.toPandas().hashes.apply(
                lambda x: pandas.Series(x).map(
                    lambda y: y.values[0])).values.astype(numpy.float32)]
        paths = save_data_models(data_np, expected, model, model_onnx,
                                 basename="SparkmlMinHashLSH")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['hashes'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import pickle
import os
import warnings
import sys
import numpy
import onnxruntime
from onnxruntime.capi.onnxruntime_pybind11_state import InvalidArgument, Fail
import pyspark
from pyspark import SparkConf
from pyspark.sql import SparkSession
from pyspark.ml.linalg import VectorUDT
from pyspark.sql.types import ArrayType, FloatType, DoubleType
from onnxmltools.utils.utils_backend import (
    compare_backend, extract_options, evaluate_condition, is_backend_enabled,
    OnnxRuntimeAssertionError, compare_outputs, ExpectedAssertionError)
from onnxmltools.utils.utils_backend_onnxruntime import _create_column


def start_spark(options):
    executable = sys.executable
    os.environ["SPARK_HOME"] = pyspark.__path__[0]
    os.environ["PYSPARK_PYTHON"] = executable
    os.environ["PYSPARK_DRIVER_PYTHON"] = executable

    builder = SparkSession.builder.appName("pyspark-unittesting").master("local[1]")
    if options:
        for k,v in options.items():
            builder.config(k, v)
    spark = builder.getOrCreate()
    # spark.sparkContext.setLogLevel("ALL")
    return spark


def stop_spark(spark):
    spark.sparkContext.stop()


def save_data_models(input, expected, model, onnx_model, basename="model", folder=None,
                     save_spark_model=False, pickle_spark_model=False, pickle_data=False):
    if folder is None:
        folder = os.environ.get('ONNXTESTDUMP', 'tests_dump')
    if not os.path.exists(folder):
        os.makedirs(folder)

    paths = []

    if pickle_spark_model:
        dest = os.path.join(folder, basename + ".expected.pkl")
        paths.append(dest)
        with open(dest, "wb") as f:
            pickle.dump(expected, f)

    if pickle_data:
        dest = os.path.join(folder, basename + ".data.pkl")
        paths.append(dest)
        with open(dest, "wb") as f:
            pickle.dump(input, f)

    if save_spark_model:
        dest = os.path.join(folder, basename + ".model")
        paths.append(dest)
        model.write().overwrite().save(dest)

    dest = os.path.join(folder, basename + ".model.onnx")
    paths.append(dest)
    with open(dest, "wb") as f:
        f.write(onnx_model.SerializeToString())
    return paths


def run_onnx_model(output_names, input, onnx_model):
    sess = onnxruntime.InferenceSession(onnx_model, providers=['CPUExecutionProvider'])
    if isinstance(input, dict):
        inputs = input
    elif isinstance(input, list):
        inp = sess.get_inputs()
        inputs = {i.name: v for i, v in zip(inp, input)}
    elif isinstance(input, numpy.ndarray):
        inp = sess.get_inputs()
        if len(inp) == 1:
            inputs = {inp[0].name: input}
        else:
            raise OnnxRuntimeAssertionError(
                "Wrong number of inputs onnx {0} != original shape {1}, onnx='{2}'".format(
                    len(inp), input.shape, onnx_model))
    else:
        raise OnnxRuntimeAssertionError(
            "Dict or list is expected, not {0}".format(type(input)))

    for k in inputs:
        if isinstance(inputs[k], list):
            inputs[k] = numpy.array(inputs[k])
    try:
        output = sess.run(output_names, inputs)
    except (InvalidArgument, Fail) as e:
        rows = []
        for inp in sess.get_inputs():
            rows.append("input: {} - {} - {}".format(inp.name, inp.type, inp.shape))
        for inp in sess.get_outputs():
            rows.append("output: {} - {} - {}".format(inp.name, inp.type, inp.shape))
        rows.append("REQUIRED: {}".format(output_names))
        for k, v in sorted(inputs.items()):
            if hasattr(v, 'shape'):
                rows.append("{}={}-{}-{}".format(k, v.shape, v.dtype, v))
            else:
                rows.append("{}={}".format(k, v))
        raise AssertionError(
            "Unable to run onnxruntime\n{}".format("\n".join(rows))) from e
        
    output_shapes = [_.shape for _ in sess.get_outputs()]
    return output, output_shapes


def compare_results(expected, output, decimal=5):
    tested = 0
    if isinstance(expected, list):
        if isinstance(output, list):
            if len(expected) != len(output):
                raise OnnxRuntimeAssertionError(
                    "Unexpected number of outputs: expected={0}, got={1}".format(len(expected), len(output)))
            for exp, out in zip(expected, output):
                compare_results(exp, out, decimal=decimal)
                tested += 1
        else:
            raise OnnxRuntimeAssertionError(
                "Type mismatch: output type is {0}".format(type(output)))
    elif isinstance(expected, dict):
        if not isinstance(output, dict):
            raise OnnxRuntimeAssertionError("Type mismatch fo")
        for k, v in output.items():
            if k not in expected:
                continue
            msg = compare_outputs(expected[k], v, decimal=decimal)
            if msg:
                raise OnnxRuntimeAssertionError("Unexpected output '{0}': \n{2}".format(k, msg))
            tested += 1
    elif isinstance(expected, numpy.ndarray):
        if isinstance(output, list):
            if expected.shape[0] == len(output) and isinstance(output[0], dict):
                import pandas
                output = pandas.DataFrame(output)
                output = output[list(sorted(output.columns))]
                output = output.values
        if isinstance(output, (dict, list)):
            if len(output) != 1:
                ex = str(output)
                if len(ex) > 70:
                    ex = ex[:70] + "..."
                raise OnnxRuntimeAssertionError(
                    "More than one output when 1 is expected\n{0}".format(ex))
            output = output[-1]
        if not isinstance(output, numpy.ndarray):
            raise OnnxRuntimeAssertionError(
                "output must be an array not {0}".format(type(output)))
        msg = compare_outputs(expected, output, decimal=decimal)
        if isinstance(msg, ExpectedAssertionError):
            raise msg
        if msg:
            raise OnnxRuntimeAssertionError(
                "Unexpected output\n{}".format(msg))
        tested += 1
    else:
        from scipy.sparse.csr import csr_matrix
        if isinstance(expected, csr_matrix):
            # DictVectorizer
            one_array = numpy.array(output)
            msg = compare_outputs(expected.todense(), one_array, decimal=decimal)
            if msg:
                raise OnnxRuntimeAssertionError("Unexpected output\n{1}".format(msg))
            tested += 1
        else:
            raise OnnxRuntimeAssertionError(
                "Unexpected type for expected output ({0})".format(type(expected)))
    if tested == 0:
        raise OnnxRuntimeAssertionError("No test for model")


def dump_data_and_sparkml_model(input, expected, model, onnx=None, basename="model", folder=None,
                        backend="onnxruntime", context=None,
                        allow_failure=None, verbose=False):
    """
    Saves data with pickle, saves the model with pickle and *onnx*,
    runs and saves the predictions for the given model.
    This function is used to test a backend (runtime) for *onnx*.

    :param input: any kind of test data
    :param expected: expected data that test results must equate to
    :param model: any model
    :param onnx: *onnx* model or *None* to use *onnxmltools* to convert it
        only if the model accepts one float vector
    :param basemodel: three files are writen ``<basename>.data.pkl``,
        ``<basename>.model.pkl``, ``<basename>.model.onnx``
    :param folder: files are written in this folder,
        it is created if it does not exist, if *folder* is None,
        it looks first in environment variable ``ONNXTESTDUMP``,
        otherwise, it is placed into ``'tests'``.
    :param backend: backend used to compare expected output and runtime output.
        Two options are currently supported: None for no test,
        `'onnxruntime'` to use module *onnxruntime*.
    :param context: used if the model contains a custom operator such
        as a custom function...
    :param allow_failure: None to raise an exception if comparison fails
        for the backends, otherwise a string which is then evaluated to check
        whether or not the test can fail, example:
        ``"StrictVersion(onnx.__version__) < StrictVersion('1.3.0')"``
    :param verbose: additional information
    :return: the created files

    Some convention for the name,
    *Bin* for a binary classifier, *Mcl* for a multiclass
    classifier, *Reg* for a regressor, *MRg* for a multi-regressor.
    The name can contain some flags. Expected outputs refer to the
    outputs computed with the original library, computed outputs
    refer to the outputs computed with a ONNX runtime.

    * ``-CannotLoad``: the model can be converted but the runtime cannot load it
    * ``-Dec3``: compares expected and computed outputs up to 3 decimals (5 by default)
    * ``-Dec4``: compares expected and computed outputs up to 4 decimals (5 by default)
    * ``-NoProb``: The original models computed probabilites for two classes *size=(N, 2)*
      but the runtime produces a vector of size *N*, the test will compare the second column
      to the column
    * ``-OneOff``: the ONNX runtime cannot computed the prediction for several inputs,
      it must be called for each of them
      and computed output.
    * ``-Out0``: only compares the first output on both sides
    * ``-Reshape``: merges all outputs into one single vector and resizes it before comparing
    * ``-SkipDim1``: before comparing expected and computed output,
      arrays with a shape like *(2, 1, 2)* becomes *(2, 2)*

    If the *backend* is not None, the function either raises an exception
    if the comparison between the expected outputs and the backend outputs
    fails or it saves the backend output and adds it to the results.
    """
    runtime_test = dict(model=model, data=input)

    if folder is None:
        folder = os.environ.get('ONNXTESTDUMP', 'tests_dump')
    if not os.path.exists(folder):
        os.makedirs(folder)

    runtime_test['expected'] = expected

    names = []
    dest = os.path.join(folder, basename + ".expected.pkl")
    names.append(dest)
    with open(dest, "wb") as f:
        pickle.dump(expected, f)

    dest = os.path.join(folder, basename + ".data.pkl")
    names.append(dest)
    with open(dest, "wb") as f:
        pickle.dump(input, f)

    dest = os.path.join(folder, basename + ".model")
    names.append(dest)
    model.write().overwrite().save(dest)

    dest = os.path.join(folder, basename + ".model.onnx")
    names.append(dest)
    with open(dest, "wb") as f:
        f.write(onnx.SerializeToString())

    runtime_test["onnx"] = dest

    # backend
    if backend is not None:
        if not isinstance(backend, list):
            backend = [backend]
        for b in backend:
            if not is_backend_enabled(b):
                continue
            if isinstance(allow_failure, str):
                allow = evaluate_condition(b, allow_failure)
            else:
                allow = allow_failure
            if allow is None:
                output = compare_backend(b, runtime_test, options=extract_options(basename),
                                         context=context, verbose=verbose)
            else:
                try:
                    output = compare_backend(b, runtime_test, options=extract_options(basename),
                                             context=context, verbose=verbose)
                except AssertionError as e:
                    if isinstance(allow, bool) and allow:
                        warnings.warn("Issue with '{0}' due to {1}".format(basename, e))
                        continue
                    else:
                        raise e
            if output is not None:
                dest = os.path.join(folder, basename + ".backend.{0}.pkl".format(b))
                names.append(dest)
                with open(dest, "wb") as f:
                    pickle.dump(output, f)

    return names


def dataframe_to_nparray(df):
    from pyspark.ml.linalg import VectorUDT
    schema = df.schema
    npcols = []
    for i in range(0, len(df.columns)):
        if isinstance(schema.fields[i].dataType, VectorUDT):
            npcols.append(df.select(df.columns[i]).toPandas().apply(
                lambda x : numpy.array(x[0].toArray())).as_matrix().reshape(-1, 1))
        else:
            npcols.append(df.select(df.columns[i]).collect())
    return numpy.array(npcols)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_converter


def convert_l2_normalization(scope, operator, container):
    # The first dimension is batch size, so the normalization is done along the 2nd axis (indexed by 1).
    attrs = {'name': operator.full_name, 'axis': 1, 'p': 2}  # Caffe normalization happens per image in one batch
    container.add_node('LpNormalization', operator.input_full_names, operator.output_full_names, **attrs)


register_converter('l2normalize', convert_l2_normalization)



# SPDX-License-Identifier: Apache-2.0

from onnxconverter_common.data_types import FloatTensorType
from .ops_names import get_sparkml_operator_name
from .ops_input_output import get_input_names, get_output_names

from ..common._container import SparkmlModelContainer
from ..common._topology import Topology

from pyspark.ml import PipelineModel


def _get_variable_for_input(scope, input_name, global_inputs, output_dict):
    '''
    Find the corresponding Variable for a given raw operator (model) name
    The variable is either supplied as graph/global inputs or has been generated as output by previous ops
    :param input_name:
    :param global_inputs:
    :param output_dict:
    :return:
    '''
    if input_name in output_dict:
        value = output_dict[input_name]
        ref_count = value[0]
        variable = value[1]
        output_dict[input_name] = [ref_count+1, variable]
        return variable

    matches = [x for x in global_inputs if x.raw_name == input_name]
    if matches:
        return matches[0]
    #
    # create a new Var
    #
    return scope.declare_local_variable(input_name)


def _parse_sparkml_simple_model(spark, scope, model, global_inputs, output_dict):
    '''
    This function handles all non-pipeline models.

    :param scope: Scope object
    :param model: A spark-ml Transformer/Evaluator (e.g., OneHotEncoder and LogisticRegression)
    :param global_inputs: A list of variables
    :param output_dict: An accumulated list of output_original_name->(ref_count, variable)
    :return: A list of output variables which will be passed to next stage
    '''
    this_operator = scope.declare_local_operator(get_sparkml_operator_name(type(model)), model)
    this_operator.raw_params = {'SparkSession': spark}
    raw_input_names = get_input_names(model)
    this_operator.inputs = [_get_variable_for_input(scope, x, global_inputs, output_dict) for x in raw_input_names]
    raw_output_names = get_output_names(model)
    for output_name in raw_output_names:
        variable = scope.declare_local_variable(output_name, FloatTensorType())
        this_operator.outputs.append(variable)
        output_dict[variable.raw_name] = [0, variable]


def _parse_sparkml_pipeline(spark, scope, model, global_inputs, output_dict):
    '''
    The basic ideas of spark-ml parsing:
        1. Sequentially go though all stages defined in the considered spark-ml pipeline
        2. The output variables of one stage will be fed into its next stage as the inputs.

    :param scope: Scope object defined in _topology.py
    :param model: spark-ml pipeline object
    :param global_inputs: A list of Variable objects
    :param output_dict: An accumulated list of output_original_name->(ref_count, variable)
    :return: A list of output variables produced by the input pipeline
    '''
    for stage in model.stages:
        _parse_sparkml(spark, scope, stage, global_inputs, output_dict)


def _parse_sparkml(spark, scope, model, global_inputs, output_dict):
    '''
    This is a delegate function. It doesn't nothing but invoke the correct parsing function according to the input
    model's type.
    :param scope: Scope object
    :param model: A spark-ml object (e.g., OneHotEncoder and LogisticRegression)
    :param inputs: A list of variables
    :return: The output variables produced by the input model
    '''
    if isinstance(model, PipelineModel):
        return _parse_sparkml_pipeline(spark, scope, model, global_inputs, output_dict)
    else:
        return _parse_sparkml_simple_model(spark, scope, model, global_inputs, output_dict)


def parse_sparkml(spark, model, initial_types=None, target_opset=None,
                  custom_conversion_functions=None, custom_shape_calculators=None):
    # Put spark-ml object into an abstract container so that our framework can work seamlessly on models created
    # with different machine learning tools.
    raw_model_container = SparkmlModelContainer(model)

    # Declare a computational graph. It will become a representation of the input spark-ml model after parsing.
    topology = Topology(raw_model_container, default_batch_size='None',
                        initial_types=initial_types,
                        target_opset=target_opset,
                        custom_conversion_functions=custom_conversion_functions,
                        custom_shape_calculators=custom_shape_calculators)

    # Declare an object to provide variables' and operators' naming mechanism. In contrast to CoreML, one global scope
    # is enough for parsing spark-ml models.
    scope = topology.declare_scope('__root__')

    # Declare input variables. They should be the inputs of the spark-ml model you want to convert into ONNX
    inputs = []
    for var_name, initial_type in initial_types:
        inputs.append(scope.declare_local_variable(var_name, initial_type))

    # The object raw_model_container is a part of the topology we're going to return. We use it to store the inputs of
    # the spark-ml's computational graph.
    for variable in inputs:
        raw_model_container.add_input(variable)

    # Parse the input spark-ml model as a Topology object.
    output_dict = {}
    _parse_sparkml(spark, scope, model, inputs, output_dict)
    outputs = []
    for k, v in output_dict.items():
        if v[0] == 0: # ref count is zero
            outputs.append(v[1])

    # THe object raw_model_container is a part of the topology we're going to return. We use it to store the outputs of
    # the spark-ml's computational graph.
    for variable in outputs:
        raw_model_container.add_output(variable)

    return topology



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter


def convert_imputer(scope, operator, container):
    op_type = 'Imputer'
    attrs = {'name': operator.full_name}
    imputer = operator.raw_operator.imputer
    if imputer.HasField('replaceDoubleValue'):
        attrs['replaced_value_float'] = imputer.replaceDoubleValue
    elif imputer.HasField('replaceInt64Value'):
        attrs['replaced_value_int64'] = imputer.replaceInt64Value
    if imputer.HasField('imputedDoubleArray'):
        attrs['imputed_value_floats'] = imputer.imputedDoubleArray.vector
    elif imputer.HasField('imputedInt64Array'):
        attrs['imputed_value_int64s'] = imputer.imputedInt64Array.vector
    container.add_node(op_type, operator.input_full_names, operator.output_full_names, op_domain='ai.onnx.ml', **attrs)


register_converter('imputer', convert_imputer)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import check_input_and_output_numbers, check_input_and_output_types
from .Convolution import calculate_convolution_and_pooling_1D_output_shape

def calculate_pooling_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. [N, C, H, W] ---> [N, C, H', W']
    '''
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    input = operator.inputs[0]
    input_shape = operator.inputs[0].type.shape

    if len(input.type.shape) != 4:
        raise RuntimeError('Input must be 4-D float tensor')

    operator.outputs[0].type.shape = [0, 0, 0, 0]
    output_shape = operator.outputs[0].type.shape

    # Adjust N-axis
    output_shape[0] = input_shape[0]

    # Adjust C-axis
    output_shape[1] = input_shape[1]

    params = operator.raw_operator.pooling
    # Set up default and non-default parameters. Notice that they are only set for H- and W-axes.
    dilations = [1, 1]  # CoreML Pooling doesn't allow dilation, so we use [1, 1] which is equivalent to no dilation.
    kernel_shape = [3, 3]
    if len(params.kernelSize) > 0:
        kernel_shape = params.kernelSize
    strides = [1, 1]
    if len(params.stride) > 0:
        strides = params.stride
    pad_mode = params.WhichOneof('PoolingPaddingType')
    if pad_mode == 'valid' and len(params.valid.paddingAmounts.borderAmounts) > 0:
        pad_amounts = params.valid.paddingAmounts.borderAmounts
        pad_heads = [pad_amounts[0].startEdgeSize, pad_amounts[1].startEdgeSize]
        pad_tails = [pad_amounts[0].endEdgeSize, pad_amounts[1].endEdgeSize]
    elif pad_mode == 'includeLastPixel' and len(params.includeLastPixel.paddingAmounts) > 0:
        pad_amounts = params.includeLastPixel.paddingAmounts
        pad_heads = [pad_amounts[0], pad_amounts[1]]
        pad_tails = [pad_amounts[0], pad_amounts[1]]
    else:
        # For same padding, padding amounts are not used
        pad_heads = [0, 0]
        pad_tails = [0, 0]

    # Calculate output shape along H- and W-axes
    for i in range(2):
        output_shape[i + 2] = calculate_convolution_and_pooling_1D_output_shape(
            input_shape[i + 2], kernel_shape[i], dilations[i], strides[i],
            pad_mode, pad_heads[i], pad_tails[i], params.globalPooling)


register_shape_calculator('pooling', calculate_pooling_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from .Scale import deduce_broadcast_axis_and_shape
from .....proto import onnx_proto
from ....common._apply_operation import apply_add, apply_reshape, apply_cast
from ....common._registration import register_converter


def convert_embedding(scope, operator, container):
    params = operator.raw_operator.embedding
    gather_op_name = scope.get_unique_operator_name('Gather')
    gather_attrs = {'name': gather_op_name}

    # Reshape the indexes we want to embed to 1-D tensor. Otherwise, ONNX Gather's output may get wrong shape.
    reshaped_input_name = scope.get_unique_variable_name(gather_op_name + 'input_reshaped')  # 2nd input of Gather
    apply_reshape(scope, operator.inputs[0].full_name, reshaped_input_name, container, desired_shape=[-1])

    # ONNX Gather accepts integers so we add a Cast to enforce this before feeding input into ONNX Gather.
    casted_input_name = scope.get_unique_variable_name(gather_op_name + 'input_casted')  # 2nd input of Gather
    apply_cast(scope, reshaped_input_name, casted_input_name, container, to=onnx_proto.TensorProto.INT64)

    # Load the embedding matrix. Its shape is outputChannels-by-inputDim.
    weights = np.array(params.weights.floatValue).reshape(params.outputChannels, params.inputDim)
    weights_name = scope.get_unique_variable_name(gather_op_name + '_W')  # 1st input of Gather
    container.add_initializer(weights_name, onnx_proto.TensorProto.FLOAT,
                              [params.inputDim, params.outputChannels], weights.transpose().flatten().tolist())

    # To support the bias term in an embedding (if exists), we need to create one extra node.
    if params.hasBias:
        # Put the embedded result onto a temporal tensor
        gather_output_name = scope.get_unique_variable_name(gather_op_name + '_output')
        container.add_node('Gather', [weights_name, casted_input_name], gather_output_name, **gather_attrs)

        # Load the bias vector into an initializer
        bias_name = scope.get_unique_variable_name(gather_op_name + '_bias')
        bias_axis, bias_shape = deduce_broadcast_axis_and_shape(container.target_opset, [params.outputChannels])
        container.add_initializer(bias_name, onnx_proto.TensorProto.FLOAT,
                                  bias_shape, params.bias.floatValue)
        # Create an addition operator to add bias (shape: [C]) into Gather's tensor (shape: [N, C])
        apply_add(scope, [gather_output_name, bias_name], operator.outputs[0].full_name, container, axis=1, broadcast=1)
    else:
        # This case has no bias, so we just output the result produced by the embedding node.
        container.add_node('Gather', [weights_name, casted_input_name], operator.output_full_names, **gather_attrs)


register_converter('embedding', convert_embedding)



# SPDX-License-Identifier: Apache-2.0

import os
from ..proto import onnx_proto
from ..convert.common import utils as convert_utils
from os import path


def load_model(file_path):
    """
    Loads an ONNX model to a ProtoBuf object.

    :param file_path: ONNX file (full file name)
    :return: ONNX model.

    Example:

    ::

        from onnxmltools.utils import load_model
        onnx_model = load_model("SqueezeNet.onnx")
    """
    if not path.exists(file_path):
        raise FileNotFoundError("{0} was not found.".format(file_path))
    model = onnx_proto.ModelProto()
    with open(file_path, 'rb') as f:
        model.ParseFromString(f.read())
    return model


def save_model(model, file_path):
    """
    Saves an ONNX model to a ProtoBuf object.
    :param model: ONNX model
    :param file_path: ONNX file (full file name)

    Example:

    ::

        from onnxmltools.utils import save_model
        save_model(onnx_model, 'c:/test_model.onnx')
    """
    if model is None or not isinstance(model, onnx_proto.ModelProto):
        raise ValueError("Model is not a valid ONNX model.")
    directory = os.path.dirname(os.path.abspath(file_path))
    if not path.exists(directory):
        raise FileNotFoundError("Directory does not exist {0}".format(directory))
    with open(file_path, 'wb') as f:
        f.write(model.SerializeToString())


def set_model_domain(model, domain):
    """
    Sets the domain on the ONNX model.

    :param model: instance of an ONNX model
    :param domain: string containing the domain name of the model

    Example:

    ::
        from onnxmltools.utils import set_model_domain
        onnx_model = load_model("SqueezeNet.onnx")
        set_model_domain(onnx_model, "com.acme")
    """
    if model is None or not isinstance(model, onnx_proto.ModelProto):
        raise ValueError("Model is not a valid ONNX model.")
    if not convert_utils.is_string_type(domain):
        raise ValueError("Domain must be a string type.")
    model.domain = domain


def set_model_version(model, version):
    """
    Sets the version of the ONNX model.

    :param model: instance of an ONNX model
    :param version: integer containing the version of the model

    Example:

    ::
        from onnxmltools.utils import set_model_version
        onnx_model = load_model("SqueezeNet.onnx")
        set_model_version(onnx_model, 1)
    """
    if model is None or not isinstance(model, onnx_proto.ModelProto):
        raise ValueError("Model is not a valid ONNX model.")
    if not convert_utils.is_numeric_type(version):
        raise ValueError("Version must be a numeric type.")
    model.model_version = version


def set_model_doc_string(model, doc, override=False):
    """
    Sets the doc string of the ONNX model.

    :param model: instance of an ONNX model
    :param doc: string containing the doc string that describes the model.
    :param override: bool if true will always override the doc string with the new value

    Example:

    ::
        from onnxmltools.utils import set_model_doc_string
        onnx_model = load_model("SqueezeNet.onnx")
        set_model_doc_string(onnx_model, "Sample doc string")
    """
    if model is None or not isinstance(model, onnx_proto.ModelProto):
        raise ValueError("Model is not a valid ONNX model.")
    if not convert_utils.is_string_type(doc):
        raise ValueError("Doc must be a string type.")
    if model.doc_string and not doc and override is False:
        raise ValueError(
            "Failing to overwrite the doc string with a blank string, set override to True if intentional.")
    model.doc_string = doc
